# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import argparse
from urllib.parse import urlsplit

import requests

from django.conf import settings
from django.db.models import F


# Wrappers around request api to ensure we always attach our user agent
# https://github.com/requests/requests/blob/master/requests/api.py


USER_AGENT = f"promgen/{settings.PROMGEN_VERSION}"
ACCEPT_HEADER = (
    "application/openmetrics-text; version=0.0.1,text/plain;version=0.0.4;q=0.5,*/*;q=0.1"
)


def post(url, data=None, json=None, **kwargs):
    headers = kwargs.setdefault("headers", {})
    headers["User-Agent"] = USER_AGENT
    return requests.post(url, data=data, json=json, **kwargs)


def get(url, params=None, **kwargs):
    headers = kwargs.setdefault("headers", {})
    headers["User-Agent"] = USER_AGENT
    return requests.get(url, params=params, **kwargs)


def delete(url, **kwargs):
    headers = kwargs.setdefault("headers", {})
    headers["User-Agent"] = USER_AGENT
    return requests.delete(url, **kwargs)


def scrape(url, params=None, **kwargs):
    """
    Scrape Prometheus target

    Light wrapper around requests.get so that we add required
    Accept headers that a target might expect
    """
    headers = kwargs.setdefault("headers", {})
    headers["Accept"] = ACCEPT_HEADER
    headers["User-Agent"] = USER_AGENT
    headers["X-Prometheus-Scrape-Timeout-Seconds"] = "10.0"
    # According to the spec, having the host with a port is optional, though
    # so by default, many clients/servers drop the port if it's known (http/https)
    # in the case of Prometheus it always forces the port in the Host header which
    # then sometimes fail for servers that do not expect it. Here we force the port
    # in the Host header to make it match how Prometheus scrapes
    # https://github.com/prometheus/prometheus/blob/2b55017379786873dc00315ffe65e22ad7026abb/scrape/target.go#L375-L387
    headers["Host"] = urlsplit(url).netloc
    return requests.get(url, params=params, **kwargs)


def setting(key, default=None, domain=None):
    """
    Settings helper based on saltstack's query

    Allows a simple way to query settings from YAML
    using the style `path:to:key` to represent

    path:
      to:
        key: value
    """
    rtn = settings.PROMGEN
    lookup = key.split(":")

    if domain:
        lookup.insert(0, domain)

    for index in lookup:
        try:
            rtn = rtn[index]
        except KeyError:
            if default != KeyError:
                return default
            raise KeyError(f"Missing required setting: {key}")
    return rtn


def inc_for_pk(model, pk, **kwargs):
    # key=F('key') + value
    model.objects.filter(pk=pk).update(**{key: F(key) + kwargs[key] for key in kwargs})


def cast(klass):
    """
    Used with argparse to cast to a Django model

    Example:
    parser.add_argument("project", type=util.cast(models.Project))
    """

    def wrapped(value):
        try:
            return klass.objects.get(name=value)
        except klass.DoesNotExist:
            raise argparse.ArgumentTypeError("Unable to find :%s" % value)

    return wrapped


def help_text(klass):
    """
    Used with argparse to lookup help_text for a Django model

    Example:
    help_text = util.help_text(models.Host)
    parser.add_argument("host", help=help_text("name"))
    """

    def wrapped(field):
        return klass._meta.get_field(field).help_text

    return wrapped


# Comment wrappers to get the docstrings from the upstream functions
get.__doc__ = requests.get.__doc__
post.__doc__ = requests.post.__doc__
delete.__doc__ = requests.delete.__doc__



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging
import warnings

from django.apps import AppConfig
from django.db.models.signals import post_migrate

logger = logging.getLogger(__name__)


def default_admin(sender, interactive, **kwargs):
    # Have to import here to ensure that the apps are already registered and
    # we get a real model instead of __fake__.User
    from django.contrib.auth.models import User

    if User.objects.filter(is_superuser=True).count() == 0:
        if interactive:
            print("  Adding default admin user")
        User.objects.create_user(
            username="admin",
            password="admin",
            is_staff=True,
            is_active=True,
            is_superuser=True,
        )
        if interactive:
            print("BE SURE TO UPDATE THE PASSWORD!!!")


def default_shard(sender, apps, interactive, **kwargs):
    Shard = apps.get_model("promgen.Shard")
    if Shard.objects.count() == 0:
        if interactive:
            print("  Adding default shard")
        Shard.objects.create(
            name="Default",
            url="http://prometheus.example.com",
            proxy=True,
            enabled=True,
        )
    if Shard.objects.filter(enabled=True).count() == 0:
        warnings.warn("No shards enabled", category=RuntimeWarning)
    if Shard.objects.filter(proxy=True).count() == 0:
        warnings.warn("No proxy shards", category=RuntimeWarning)


def check_site(app_config, **kwargs):
    Site = app_config.get_model("Site")
    if Site.objects.filter(domain__in=["example.com"]).count():
        warnings.warn("Site unconfigured", category=RuntimeWarning)


class PromgenConfig(AppConfig):
    name = "promgen"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from promgen import celery, checks, signals  # NOQA

        post_migrate.connect(default_shard, sender=self)
        post_migrate.connect(default_admin, sender=self)
        post_migrate.connect(check_site, sender=self)



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from unittest import mock

from requests.exceptions import RequestException

from django.test import override_settings

from promgen import models, rest, tests
from promgen.notification.webhook import NotificationWebhook


class WebhookTest(tests.PromgenTest):
    def setUp(self):
        one = models.Project.objects.get(pk=1)
        two = models.Service.objects.get(pk=1)

        self.senderA = NotificationWebhook.create(
            obj=one, value="http://webhook.example.com/project"
        )
        self.senderB = NotificationWebhook.create(
            obj=two, value="http://webhook.example.com/service"
        )

    @override_settings(PROMGEN=tests.SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    @override_settings(CELERY_TASK_EAGER_PROPAGATES=True)
    @mock.patch("promgen.util.post")
    def test_webhook(self, mock_post):
        response = self.fireAlert()
        self.assertRoute(response, rest.AlertReceiver, 202)
        self.assertCount(models.AlertError, 0, "No failed alerts")

        self.assertCount(models.Alert, 1, "Alert should be queued")
        self.assertEqual(mock_post.call_count, 2, "Two alerts should be sent")

        # Our sample is the same as the original, with some annotations added
        _SAMPLE = tests.Data("notification", "webhook.json").json()
        # External URL is depended on test order
        _SAMPLE["externalURL"] = mock.ANY

        mock_post.assert_has_calls(
            [
                mock.call("http://webhook.example.com/project", json=_SAMPLE),
                mock.call("http://webhook.example.com/service", json=_SAMPLE),
            ],
            any_order=True,
        )

    @override_settings(PROMGEN=tests.SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    @override_settings(CELERY_TASK_EAGER_PROPAGATES=True)
    @mock.patch("promgen.util.post")
    def test_filter(self, mock_post):
        # Our first sender will only allow critical messages
        self.senderA.filter_set.create(name="severity", value="critical")

        # Our second sender allows critical and major
        self.senderB.filter_set.create(name="severity", value="critical")
        self.senderB.filter_set.create(name="severity", value="major")

        self.assertCount(models.Filter, 3, "Should be three filters")

        response = self.fireAlert()
        self.assertRoute(response, rest.AlertReceiver, 202)
        self.assertCount(models.AlertError, 0, "No failed alerts")
        self.assertCount(models.Alert, 1, "Alert should be queued")
        self.assertEqual(mock_post.call_count, 1, "One notification should be skipped")

    @override_settings(PROMGEN=tests.SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    @override_settings(CELERY_TASK_EAGER_PROPAGATES=True)
    @mock.patch("promgen.util.post")
    def test_failure(self, mock_post):
        # When our post results in a failure, then our error_count should be
        # properly updated and some errors should be logged to be viewed later
        mock_post.side_effect = RequestException("Boom!")

        response = self.fireAlert()
        self.assertRoute(response, rest.AlertReceiver, 202)
        self.assertCount(models.Alert, 1, "Alert should be queued")
        self.assertCount(models.AlertError, 2, "Two errors should be logged")
        self.assertEqual(mock_post.call_count, 2, "Two posts should be attempted")

        alert = models.Alert.objects.first()
        self.assertEqual(alert.sent_count, 0, "No successful sent")
        self.assertEqual(alert.error_count, 2, "Error incremented")



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import re

from dateutil import parser

from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator, URLValidator

# See definition of duration field
# https://prometheus.io/docs/prometheus/latest/configuration/configuration/#configuration-file

duration = RegexValidator(
    r"[0-9]+(ms|[smhdwy])",
    "Invalid or missing duration suffix. Example: 30s, 5m, 1h ([0-9]+(ms|[smhdwy])",
)

# Label Value Definition
# https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels
metricname = RegexValidator(
    r"^[a-zA-Z_:][a-zA-Z0-9_:]*$",
    "Only alphanumeric characters are allowed.",
)
labelname = RegexValidator(
    r"^[a-zA-Z_][a-zA-Z0-9_]*$",
    "Only alphanumeric characters are allowed.",
)

# While Prometheus accepts label values of any unicode character, our values sometimes
# make it into URLs, so we want to make sure we do not have stray / characters
labelvalue = RegexValidator(
    r"^[\w][- \w]+$", "Unicode letters, numbers, underscores, or hyphens or spaces"
)

hostname = RegexValidator(
    regex=r"^("
    + URLValidator.ipv4_re
    + "|"
    + URLValidator.ipv6_re
    + "|"
    + URLValidator.host_re
    + "|"
    + URLValidator.hostname_re
    + ")$",
    message="Invalid hostname: %(value)s",
    flags=re.IGNORECASE,
)


def datetime(value):
    try:
        parser.parse(value)
    except ValueError:
        raise ValidationError("Invalid timestamp")



# Generated by Django 2.2.10 on 2020-02-21 05:56

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0014_exporter_scheme'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rule',
            name='content_type',
            field=models.ForeignKey(limit_choices_to=models.Q(models.Q(('app_label', 'promgen'), ('model', 'site')), models.Q(('app_label', 'promgen'), ('model', 'project')), models.Q(('app_label', 'promgen'), ('model', 'service')), _connector='OR'), on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType'),
        ),
        migrations.AlterField(
            model_name='sender',
            name='content_type',
            field=models.ForeignKey(limit_choices_to=models.Q(models.Q(('app_label', 'auth'), ('model', 'user')), models.Q(('app_label', 'promgen'), ('model', 'project')), models.Q(('app_label', 'promgen'), ('model', 'service')), _connector='OR'), on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType'),
        ),
    ]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging
from functools import wraps

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.models import Group
from django.core.cache import cache
from django.db.models.signals import post_delete, post_save, pre_delete, pre_save
from django.dispatch import Signal, receiver
from promgen import models, prometheus, tasks

logger = logging.getLogger(__name__)

trigger_write_config = Signal()
trigger_write_rules = Signal()
trigger_write_urls = Signal()
post_reload = Signal()


def multi_receiver(signal, senders, **kwargs):
    def _decorator(func):
        for sender in senders:
            signal.connect(func, sender=sender, **kwargs)
        return func

    return _decorator


def run_once(signal):
    """
    Run a signal only once

    Certain actions we want to run only once, at the end of
    processing so we wrap our function in a special decorator
    that uses Django's caching system to set whether we
    want to run it or not, and trigger the actual run with
    a force keyword at the end of the request when we run to run it
    """

    def _decorator(func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            key = f"{func.__module__}.{func.__name__}"
            if "force" in kwargs:
                logger.debug("Checking %s for %s", key, kwargs["sender"])
                kwargs.pop("force")
                if cache.get(key):
                    cache.delete(key)
                    logger.debug("Running %s for %s", key, kwargs["sender"])
                    return func(*args, **kwargs)
            else:
                logger.debug("Queueing %s for %s", key, kwargs["sender"])
                cache.set(key, 1)

        signal.connect(_wrapper)
        return _wrapper

    return _decorator


def skip_raw(func):
    """
    For many of our signals that call out to an external service, we want to skip
    it any time we have a raw object from a fixture. This decorator helps us centralize
    our check code to keep it consistent
    """

    @wraps(func)
    def _wrapper(*, raw=False, instance, **kwargs):
        if raw:
            logger.debug("Skipping %s:%s for raw %s", __name__, func.__name__, instance)
            return
        logger.debug("Running %s:%s for %s", __name__, func.__name__, instance)
        return func(raw=raw, instance=instance, **kwargs)

    return _wrapper


@run_once(trigger_write_config)
def _trigger_write_config(signal, **kwargs):
    targets = [server.host for server in models.Prometheus.objects.all()]
    for target in targets:
        logger.info("Queueing write_config on %s", target)
        tasks.write_config.apply_async(queue=target)
    if "request" in kwargs:
        messages.info(kwargs["request"], f"Updating config on {targets}")
    return True


@run_once(trigger_write_rules)
def _trigger_write_rules(signal, **kwargs):
    targets = [server.host for server in models.Prometheus.objects.all()]
    for target in targets:
        logger.info("Queueing write_rules on %s", target)
        tasks.write_rules.apply_async(queue=target)
    if "request" in kwargs:
        messages.info(kwargs["request"], f"Updating rules on {targets}")
    return True


@run_once(trigger_write_urls)
def _trigger_write_urls(signal, **kwargs):
    targets = [server.host for server in models.Prometheus.objects.all()]
    for target in targets:
        logger.info("Queueing write_urls on %s", target)
        tasks.write_urls.apply_async(queue=target)
    if "request" in kwargs:
        messages.info(kwargs["request"], f"Updating urls on {targets}")
    return True


@skip_raw
def update_log(sender, instance, **kwargs):
    # For our update_log, we hook the pre_save signal and make sure it's an
    # existing object by checking for a primary key. We then use that to get a
    # copy of the existing object from the database so that we can show the
    # changes
    if instance.pk:
        old = sender.objects.get(pk=instance.pk)
        models.Audit.log(f"Updated {sender.__name__} {instance}", instance, old)


pre_save.connect(update_log, sender=models.Exporter)
pre_save.connect(update_log, sender=models.Farm)
pre_save.connect(update_log, sender=models.Host)
pre_save.connect(update_log, sender=models.Project)
pre_save.connect(update_log, sender=models.Rule)
pre_save.connect(update_log, sender=models.Sender)
pre_save.connect(update_log, sender=models.Service)
pre_save.connect(update_log, sender=models.URL)


@skip_raw
def create_log(sender, instance, created, **kwargs):
    # For our create_log, we have to hook post_save to make sure we have a
    # primary key set so that we can link back to it using the ContentType
    # system.
    if created:
        models.Audit.log(f"Created {sender.__name__} {instance}", instance)


post_save.connect(create_log, sender=models.Exporter)
post_save.connect(create_log, sender=models.Farm)
post_save.connect(create_log, sender=models.Host)
post_save.connect(create_log, sender=models.Project)
post_save.connect(create_log, sender=models.Rule)
post_save.connect(create_log, sender=models.Sender)
post_save.connect(create_log, sender=models.Service)
post_save.connect(create_log, sender=models.URL)


def delete_log(sender, instance, **kwargs):
    models.Audit.log(f"Deleted {sender.__name__} {instance}", instance)


post_delete.connect(delete_log, sender=models.Exporter)
post_delete.connect(delete_log, sender=models.Farm)
post_delete.connect(delete_log, sender=models.Host)
post_delete.connect(delete_log, sender=models.Project)
post_delete.connect(delete_log, sender=models.Rule)
post_delete.connect(delete_log, sender=models.Sender)
post_delete.connect(delete_log, sender=models.Service)
post_delete.connect(delete_log, sender=models.URL)


@receiver(post_save, sender=models.Rule)
@skip_raw
def save_rule(sender, instance, **kwargs):
    prometheus.check_rules([instance])
    trigger_write_rules.send(instance)


@receiver(post_delete, sender=models.Rule)
def delete_rule(sender, instance, **kwargs):
    trigger_write_rules.send(instance)


@receiver(post_save, sender=models.URL)
@skip_raw
def save_url(sender, instance, **kwargs):
    trigger_write_urls.send(instance)


@receiver(post_delete, sender=models.URL)
def delete_url(sender, instance, **kwargs):
    trigger_write_urls.send(instance)


@receiver(post_save, sender=models.Host)
@skip_raw
def save_host(sender, instance, **kwargs):
    """Only trigger write if parent project also has exporters"""
    for project in instance.farm.project_set.all():
        if project.exporter_set:
            trigger_write_config.send(instance)


@receiver(pre_delete, sender=models.Host)
def delete_host(sender, instance, **kwargs):
    """Only trigger write if parent project also has exporters"""
    for project in instance.farm.project_set.all():
        if project.exporter_set.exists():
            trigger_write_config.send(instance)


@receiver(pre_delete, sender=models.Farm)
def delete_farm(sender, instance, **kwargs):
    """Only trigger write if parent project also has exporters"""
    for project in instance.project_set.all():
        trigger_write_config.send(instance)


@receiver(post_save, sender=models.Exporter)
@skip_raw
def save_exporter(sender, instance, **kwargs):
    """Only trigger write if parent project also has hosts"""
    if instance.project.farm:
        if instance.project.farm.host_set.exists():
            trigger_write_config.send(instance)


@receiver(pre_delete, sender=models.Exporter)
def delete_exporter(sender, instance, **kwargs):
    """Only trigger write if parent project also has hosts"""
    if instance.project.farm:
        if instance.project.farm.host_set.exists():
            trigger_write_config.send(instance)


@receiver(post_save, sender=models.Project)
@skip_raw
def save_project(instance, **kwargs):
    if instance.farm and instance.farm.host_set.exists() and instance.exporter_set.exists():
        trigger_write_config.send(instance)
        return True


@receiver(pre_delete, sender=models.Project)
def delete_project(sender, instance, **kwargs):
    if instance.farm and instance.farm.host_set.exists() and instance.exporter_set.exists():
        trigger_write_config.send(instance)


@receiver(post_save, sender=models.Service)
@skip_raw
def save_service(*, sender, instance, **kwargs):
    # We saving a service, we delegate the configuration reload triggering to
    # the child projects which have additional information about if we need to
    # write out our file or not. We call our save_project signal directly
    # (instead of through post_save.save) because we don't want to trigger other
    # attached signals
    # We don't use sender here, but put it in our parameters so we don't pass
    # two sender entries to save_project
    for project in instance.project_set.prefetch_related("farm", "farm__host_set", "exporter_set"):
        if save_project(sender=models.Project, instance=project, **kwargs):
            # If any of our save_project returns True, then we do not need to
            # check any others
            return True


@receiver(post_save, sender=settings.AUTH_USER_MODEL)
@skip_raw
def add_user_to_default_group(instance, created, **kwargs):
    # If we enabled our default group, then we want to ensure that all newly
    # created users are also added to our default group so they inherit the
    # default permissions
    if not settings.PROMGEN_DEFAULT_GROUP:
        return
    if not created:
        return

    instance.groups.add(Group.objects.get(name=settings.PROMGEN_DEFAULT_GROUP))


@receiver(post_save, sender=settings.AUTH_USER_MODEL)
@skip_raw
def add_email_sender(instance, created, **kwargs):
    if instance.email:
        models.Sender.objects.get_or_create(
            obj=instance, sender="promgen.notification.email", value=instance.email
        )
    else:
        logger.warning("No email for user %s", instance)


# Not a 'real' signal but we match most of the interface for post_save
def check_user_subscription(sender, instance, created, request):
    # When a user subscribes to a notification, we want to ensure
    # they have a default notification enabled
    if instance.sender != "promgen.notification.user":
        messages.success(request, "Subscription saved")
        return

    notifiers = models.Sender.objects.filter(obj=instance.owner)
    if notifiers:
        logger.debug("Existing notifiers found")
        return

    if instance.owner.email:
        models.Sender.objects.get_or_create(
            obj=instance.owner,
            sender="promgen.notification.email",
            value=instance.owner.email,
        )
        messages.success(request, "Subscribed using %s" % (instance.owner.email))
    else:
        messages.warning(request, "No email configured")


@receiver(post_save, sender=models.Service)
def add_default_service_subscription(instance, created, **kwargs):
    if created and instance.owner:
        sender, new_notifier = models.Sender.objects.get_or_create(
            obj=instance,
            sender="promgen.notification.user",
            value=instance.owner.username,
            defaults={"owner": instance.owner},
        )


@receiver(post_save, sender=models.Project)
def add_default_project_subscription(instance, created, **kwargs):
    if created and instance.owner:
        sender, new_notifier = models.Sender.objects.get_or_create(
            obj=instance,
            sender="promgen.notification.user",
            value=instance.owner.username,
            defaults={"owner": instance.owner},
        )



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

"""
Simple webhook bridge
Accepts alert json from Alert Manager and then POSTs individual alerts to
configured webhook destinations
"""

import logging

from django import forms

from promgen import util
from promgen.notification import NotificationBase

logger = logging.getLogger(__name__)


class FormWebhook(forms.Form):
    value = forms.URLField(
        required=True,
        label="URL",
    )
    alias = forms.CharField(
        required=False,
        help_text="Optional description to be displayed instead of the URL.",
    )


class NotificationWebhook(NotificationBase):
    """
    Post notifications to a specific web endpoint.
    """

    form = FormWebhook

    def _send(self, url, data):
        util.post(url, json=data).raise_for_status()



# Generated by Django 2.1.8 on 2019-05-31 04:46

from django.db import migrations, models
import django.db.models.deletion


def default_probe(apps, schema_editor):
    Probe = apps.get_model("promgen", "Probe")
    Probe.objects.get_or_create(
        pk=1, defaults={"module": "http_2xx", "description": "Default http check"}
    )


class Migration(migrations.Migration):

    dependencies = [("promgen", "0008_shard_service")]

    operations = [
        migrations.CreateModel(
            name="Probe",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "module",
                    models.CharField(
                        max_length=128,
                        help_text="Probe Module from blackbox_exporter config",
                        unique=True,
                    ),
                ),
                ("description", models.TextField(blank=True)),
            ],
        ),
        migrations.RunPython(default_probe, migrations.RunPython.noop),
        migrations.AddField(
            model_name="url",
            name="probe",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                to="promgen.Probe",
            ),
            preserve_default=False,
        ),
    ]



# These sources are released under the terms of the MIT license: see LICENSE

import logging

from django import forms

from promgen import util
from promgen.notification import NotificationBase

logger = logging.getLogger(__name__)


class FormSlack(forms.Form):
    value = forms.URLField(
        required=True,
        label="Slack Incoming Webhook URL",
    )
    alias = forms.CharField(
        required=False,
        help_text="Optional description to be displayed instead of the URL.",
    )


class NotificationSlack(NotificationBase):
    """
    Send messages to Slack via Incoming Webhook.

    An Incoming Webhook has to be configured for your workspace; you
    can set one up here:

    https://my.slack.com/services/new/incoming-webhook/

    A fitting Prometheus icon can be selected from here:

    https://github.com/quintessence/slack-icons
    """

    form = FormSlack

    def _send(self, url, data):
        kwargs = {}
        proxy = self.config("proxies", default=None)
        if proxy:
            kwargs["proxies"] = {
                "http": proxy,
                "https": proxy,
            }

        if data["status"] == "resolved":
            message = self.render("promgen/sender/slack.resolved.txt", data)
        else:
            message = self.render("promgen/sender/slack.body.txt", data)

        json = {
            "text": message,
        }

        util.post(url, json=json, **kwargs).raise_for_status()



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE


FARM_DEFAULT = "promgen"


class DiscoveryBase:
    remote = True

    """
    Basic discovery plugin base

    Child classes should implement both fetch and farm methods
    """

    def fetch(self, farm):
        """
        Return list of hosts for farm
        """
        raise NotImplementedError()

    def farms(self):
        """
        Return a list of farm names
        """
        raise NotImplementedError()



# Generated by Django 3.2.13 on 2023-07-11 03:02

from django.db import migrations, models


def forward(apps, schema_editor):
    Rule = apps.get_model("promgen", "Rule")
    Label = apps.get_model("promgen", "RuleLabel")
    Annotation = apps.get_model("promgen", "RuleAnnotation")

    # For our forward migration, we want to loop through all the old Label and Annotation entries
    # and convert them to a dictionary property on our Rule model
    for rule in Rule.objects.all():
        rule.labels = {l.name: l.value for l in Label.objects.filter(rule_id=rule.id)}
        rule.annotations = {l.name: l.value for l in Annotation.objects.filter(rule_id=rule.id)}
        rule.save()


def reverse(apps, schema_editor):
    Rule = apps.get_model("promgen", "Rule")
    Label = apps.get_model("promgen", "RuleLabel")
    Annotation = apps.get_model("promgen", "RuleAnnotation")
    for rule in Rule.objects.all():
        Label.objects.bulk_create(
            [Label(rule_id=rule.id, name=x, value=rule.labels[x]) for x in rule.labels]
        )
        Annotation.objects.bulk_create(
            [Annotation(rule_id=rule.id, name=x, value=rule.annotations[x]) for x in rule.annotations]
        )


class Migration(migrations.Migration):

    dependencies = [
        ("promgen", "0021_shard_load"),
    ]

    operations = [
        migrations.AddField(
            model_name="rule",
            name="annotations",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="rule",
            name="labels",
            field=models.JSONField(default=dict),
        ),
        migrations.RunPython(forward, reverse),
        migrations.DeleteModel(
            name="RuleAnnotation",
        ),
        migrations.DeleteModel(
            name="RuleLabel",
        ),
    ]



# Generated by Django 2.2.10 on 2020-03-31 05:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0019_sender_enabled'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='exporter',
            unique_together={('job', 'port', 'path', 'scheme', 'project')},
        ),
    ]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging

from django import forms

from promgen import util
from promgen.notification import NotificationBase

logger = logging.getLogger(__name__)


class FormLineNotify(forms.Form):
    value = forms.CharField(
        required=True,
        label="LINE Notify Token",
    )
    alias = forms.CharField(
        required=True,
        help_text="Use to hide token from being displayed.",
    )


class NotificationLineNotify(NotificationBase):
    """
    Send messages to LINE Notify.

    https://notify-bot.line.me/en/
    """

    form = FormLineNotify

    def _send(self, token, data):
        url = self.config("server")

        if data["status"] == "resolved":
            message = self.render("promgen/sender/linenotify.resolved.txt", data)
        else:
            message = self.render("promgen/sender/linenotify.body.txt", data)

        params = {
            "message": message,
        }

        headers = {"Authorization": "Bearer %s" % token}

        util.post(url, data=params, headers=headers).raise_for_status()



# Copyright (c) 2019 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import concurrent.futures
import json
import logging
from http import HTTPStatus
from urllib.parse import urljoin

import requests
from django.http import HttpResponse, JsonResponse
from django.views.generic import View
from django.views.generic.base import TemplateView
from promgen import forms, models, prometheus, util
from requests.exceptions import HTTPError

logger = logging.getLogger(__name__)


def proxy_error(response):
    """
    Return a wrapped proxy error

    Taking a request.response object as input, return it slightly modified
    with an extra header for debugging so that we can see where the request
    failed
    """
    r = HttpResponse(
        response.content,
        content_type=response.headers["content-type"],
        status=response.status_code,
    )
    r.setdefault("X-PROMGEN-PROXY", response.url)
    return r


class PrometheusProxy(View):
    proxy_headers = {"HTTP_REFERER": "Referer"}

    @property
    def headers(self):
        # Loop through the headers from our request, and decide which ones
        # we should pass through upstream. Currently, our 'Referer' header is
        # the main one we are interested in, since this can help us debug which
        # grafana dashboard is responsible for the query.
        return {
            self.proxy_headers[k]: self.request.META[k]
            for k in self.proxy_headers
            if k in self.request.META
        }

    def proxy(self, request):
        futures = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            for host in models.Shard.objects.filter(proxy=True):
                futures.append(
                    executor.submit(
                        util.get,
                        urljoin(host.url, request.get_full_path_info()),
                        headers=self.headers,
                    )
                )
            yield from concurrent.futures.as_completed(futures)


class ProxyGraph(TemplateView):
    template_name = "promgen/graph.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["shard_list"] = models.Shard.objects.filter(proxy=True)
        for k, v in self.request.GET.items():
            _, k = k.split(".")
            context[k] = v
        return context


class ProxyLabels(PrometheusProxy):
    def get(self, request):
        data = set()
        for future in self.proxy(request):
            try:
                result = future.result()
                result.raise_for_status()
                _json = result.json()
                logger.debug("Appending data from %s", result.url)
                data.update(_json["data"])
            except HTTPError:
                logger.warning("Error with response")
                return proxy_error(result)

        return JsonResponse({"status": "success", "data": sorted(data)})


class ProxyLabelValues(PrometheusProxy):
    def get(self, request, label):
        data = set()
        for future in self.proxy(request):
            try:
                result = future.result()
                result.raise_for_status()
                _json = result.json()
                logger.debug("Appending data from %s", result.url)
                data.update(_json["data"])
            except HTTPError:
                logger.warning("Error with response")
                return proxy_error(result)

        return JsonResponse({"status": "success", "data": sorted(data)})


class ProxySeries(PrometheusProxy):
    def get(self, request):
        data = []
        for future in self.proxy(request):
            try:
                result = future.result()
                result.raise_for_status()
                _json = result.json()
                logger.debug("Appending data from %s", result.url)
                data += _json["data"]
            except HTTPError:
                logger.warning("Error with response")
                return proxy_error(result)

        return JsonResponse({"status": "success", "data": data})


class ProxyQueryRange(PrometheusProxy):
    def get(self, request):
        data = []
        resultType = None
        for future in self.proxy(request):
            try:
                result = future.result()
                result.raise_for_status()
                _json = result.json()
                logger.debug("Appending data from %s", result.url)
                data += _json["data"]["result"]
                resultType = _json["data"]["resultType"]
            except HTTPError:
                return proxy_error(result)

        return JsonResponse(
            {"status": "success", "data": {"resultType": resultType, "result": data}}
        )


class ProxyQuery(PrometheusProxy):
    def get(self, request):
        data = []
        resultType = None
        for future in self.proxy(request):
            try:
                result = future.result()
                result.raise_for_status()
                _json = result.json()
                logger.debug("Appending data from %s", result.url)
                data += _json["data"]["result"]
                resultType = _json["data"]["resultType"]
            except HTTPError:
                return proxy_error(result)

        return JsonResponse(
            {"status": "success", "data": {"resultType": resultType, "result": data}}
        )


class ProxyAlerts(View):
    def get(self, request):
        try:
            url = urljoin(util.setting("alertmanager:url"), "/api/v2/alerts")
            response = util.get(url)
        except requests.exceptions.ConnectionError:
            logger.error("Error connecting to %s", url)
            return JsonResponse({}, status=HTTPStatus.INTERNAL_SERVER_ERROR)
        else:
            return HttpResponse(response.content, content_type="application/json")


class ProxySilences(View):
    def get(self, request):
        try:
            url = urljoin(util.setting("alertmanager:url"), "/api/v2/silences")
            response = util.get(url, params={"silenced": False})
        except requests.exceptions.ConnectionError:
            logger.error("Error connecting to %s", url)
            return JsonResponse({}, status=HTTPStatus.INTERNAL_SERVER_ERROR)
        else:
            return HttpResponse(response.content, content_type="application/json")

    def post(self, request):
        body = json.loads(request.body.decode("utf-8"))
        body.setdefault("comment", "Silenced from Promgen")
        body.setdefault("createdBy", request.user.email)

        form = forms.SilenceForm(body)
        if not form.is_valid():
            return JsonResponse(
                {
                    "messages": [
                        {"class": "alert alert-warning", "message": m, "label": k}
                        for k in form.errors
                        for m in form.errors[k]
                    ]
                },
                status=HTTPStatus.UNPROCESSABLE_ENTITY,
            )

        try:
            response = prometheus.silence(**form.cleaned_data)
        except requests.HTTPError as e:
            return JsonResponse(
                {
                    "messages": [
                        {"class": "alert alert-danger", "message": e.response.text}
                    ]
                },
                status=e.response.status_code,
            )
        except Exception as e:
            return JsonResponse(
                {"messages": [{"class": "alert alert-danger", "message": str(e)}]},
                status=HTTPStatus.UNPROCESSABLE_ENTITY,
            )

        return HttpResponse(
            response.text, status=response.status_code, content_type="application/json"
        )


class ProxyDeleteSilence(View):
    def delete(self, request, silence_id):
        url = urljoin(util.setting("alertmanager:url"), f"/api/v2/silence/{silence_id}")
        response = util.delete(url)
        return HttpResponse(
            response.text, status=response.status_code, content_type="application/json"
        )



# Generated by Django 2.1.8 on 2019-04-22 07:15

from django.db import migrations, models
import django.db.models.deletion


def project_to_shard(apps, schema_editor):
    Project = apps.get_model("promgen", "Project")
    for project in Project.objects.all():
        project.shard_id = project.service.shard_id
        project.save()


def project_to_service(apps, schema_editor):
    Project = apps.get_model("promgen", "Project")
    for project in Project.objects.all():
        project.service.shard_id = project.shard_id
        project.service.save()


class Migration(migrations.Migration):

    dependencies = [("promgen", "0007_message_filter")]

    operations = [
        migrations.AddField(
            model_name="project",
            name="shard",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                to="promgen.Shard",
            ),
            preserve_default=False,
        ),
        migrations.RunPython(project_to_shard, project_to_service),
        migrations.AlterModelOptions(name="service", options={"ordering": ["name"]}),
        migrations.RemoveField(model_name="service", name="shard"),
    ]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from unittest import mock

from django.test import override_settings

from promgen import models, rest, tests
from promgen.notification.email import NotificationEmail


class EmailTest(tests.PromgenTest):
    def setUp(self):
        one = models.Project.objects.get(pk=1)
        two = models.Project.objects.get(pk=2)

        NotificationEmail.create(obj=one, value="example@example.com")
        NotificationEmail.create(obj=one, value="foo@example.com")
        NotificationEmail.create(obj=two, value="bar@example.com")

    @override_settings(PROMGEN=tests.SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    @mock.patch("promgen.notification.email.send_mail")
    def test_email(self, mock_email):
        response = self.fireAlert()
        self.assertRoute(response, rest.AlertReceiver, 202)
        self.assertCount(models.AlertError, 0, "No failed alerts")

        _SUBJECT = tests.Data("notification", "email.subject.txt").raw().strip()
        _MESSAGE = tests.Data("notification", "email.body.txt").raw().strip()

        mock_email.assert_has_calls(
            [
                mock.call(_SUBJECT, _MESSAGE, "promgen@example.com", ["example@example.com"]),
                mock.call(_SUBJECT, _MESSAGE, "promgen@example.com", ["foo@example.com"]),
            ],
            any_order=True,
        )
        # Three senders are registered but only two should trigger
        self.assertTrue(mock_email.call_count == 2)



# Copyright (c) 2022 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE
from django.urls import reverse

from promgen import tests, views


class WebTests(tests.PromgenTest):
    fixtures = ["testcases.yaml", "extras.yaml"]

    route_map = [
        # viewname, viewclass, extra params
        ("datasource-list", views.DatasourceList, {}),
        ("datasource-detail", views.DatasourceDetail, {"pk": 1}),
        ("service-list", views.ServiceList, {}),
        ("service-detail", views.ServiceDetail, {"pk": 1}),
        ("project-detail", views.ProjectDetail, {"pk": 1}),
        ("farm-link", views.FarmLink, {"pk": 1, "source": "promgen"}),
        ("project-exporter", views.ExporterRegister, {"pk": 1}),
        ("project-notifier", views.ProjectNotifierRegister, {"pk": 1}),
        ("url-list", views.URLList, {}),
        ("farm-list", views.FarmList, {}),
        ("farm-detail", views.FarmDetail, {"pk": 1}),
        ("host-list", views.HostList, {}),
        ("host-detail", views.HostDetail, {"slug": "example.com"}),
        ("rules-list", views.RulesList, {}),
        ("rule-detail", views.RuleDetail, {"pk": 1}),
        ("audit-list", views.AuditList, {}),
        ("site-detail", views.SiteDetail, {}),
        ("profile", views.Profile, {}),
        # For this test, we're testing a non-admin user
        # so we expect this page to redirect to 302
        ("import", views.Import, {"status_code": 302}),
        ("alert-list", views.AlertList, {}),
        ("alert-detail", views.AlertDetail, {"pk": 1}),
    ]

    def setUp(self):
        self.user = self.force_login(username="demo")

    def test_routes(self):
        for viewname, viewclass, params in self.route_map:
            # By default we'll pass all params as-is to our reverse()
            # method, but we may have a few special ones (like status_code)
            # that we want to pop and handle separately
            # Default to checking for a 200 unless we need to specifically
            # check for a redirect or some other status code.
            status_code = params.pop("status_code", 200)
            with self.subTest(viewname=viewname, params=params):
                response = self.client.get(reverse(viewname, kwargs=params))
                self.assertRoute(response, viewclass, status_code)



# Generated by Django 2.2.10 on 2020-02-18 09:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0013_validation_fix'),
    ]

    operations = [
        migrations.AddField(
            model_name='exporter',
            name='scheme',
            field=models.CharField(choices=[('http', 'http'), ('https', 'https')], default='http', help_text='Scrape exporter over http or https', max_length=5),
        ),
        migrations.AlterField(
            model_name='exporter',
            name='job',
            field=models.CharField(help_text='Exporter name. Example node, jmx, app', max_length=128),
        ),
        migrations.AlterField(
            model_name='exporter',
            name='path',
            field=models.CharField(blank=True, help_text='Exporter path. Defaults to /metrics', max_length=128),
        ),
        migrations.AlterField(
            model_name='exporter',
            name='port',
            field=models.IntegerField(help_text='Port Exporter is running on'),
        ),
    ]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from urllib.parse import urlunsplit

from django.conf import settings
from django.shortcuts import resolve_url
from promgen import models


def resolve_domain(*args, **kwargs):
    return urlunsplit(
        (
            settings.PROMGEN_SCHEME,
            models.Site.objects.get_current().domain,
            resolve_url(*args, **kwargs),
            "",
            "",
        )
    )



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from unittest import mock

from django.test import override_settings

from promgen import models, rest, tests
from promgen.notification.email import NotificationEmail
from promgen.notification.linenotify import NotificationLineNotify
from promgen.notification.user import NotificationUser


class UserSplayTest(tests.PromgenTest):
    @override_settings(PROMGEN=tests.SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    @mock.patch("promgen.notification.email.send_mail")
    @mock.patch("promgen.util.post")
    def test_user_splay(self, mock_email, mock_post):
        one = models.Service.objects.get(pk=1)

        NotificationUser.create(obj=one, value=one.owner.username)
        NotificationLineNotify.create(obj=one.owner, value="#foo")
        NotificationEmail.create(obj=one.owner, value="foo@bar.example")

        response = self.fireAlert()
        self.assertRoute(response, rest.AlertReceiver, 202)
        self.assertCount(models.Alert, 1, "Alert Queued")
        self.assertCount(models.AlertError, 0, "No failed alerts")

        # Since we test the specifics elsewhere, just want to check
        # the count of calls here
        self.assertEqual(mock_post.call_count, 1, "Called LINE Notify")
        self.assertEqual(mock_email.call_count, 1, "Called email")

    @override_settings(PROMGEN=tests.SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    @mock.patch("promgen.notification.email.send_mail")
    def test_failed_user(self, mock_email):
        # We have one valid sender and one invalid one
        # The invalid one should be skipped while still letting
        # the valid one pass
        one = models.Service.objects.get(pk=1)
        NotificationEmail.create(obj=one, value="foo@bar.example")
        NotificationUser.create(obj=one, value="does not exist")

        response = self.fireAlert()
        self.assertRoute(response, rest.AlertReceiver, 202)
        self.assertCount(models.Alert, 1, "Alert Queued")
        self.assertCount(models.AlertError, 0, "No failed alerts")

        self.assertEqual(mock_email.call_count, 1, "Still called email")



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE
import json

from django import forms
from django.contrib import admin
from django.utils.html import format_html

from promgen import actions, models, plugins


class PrometheusInline(admin.TabularInline):
    model = models.Prometheus


class FilterInline(admin.TabularInline):
    model = models.Filter


@admin.register(models.Host)
class HostAdmin(admin.ModelAdmin):
    list_display = ("name", "farm")


@admin.register(models.Shard)
class ShardAdmin(admin.ModelAdmin):
    list_display = ("name", "url", "proxy", "enabled")
    list_filter = ("proxy", "enabled")
    inlines = [PrometheusInline]
    actions = [
        actions.shard_targets,
        actions.shard_rules,
        actions.shard_urls,
    ]


@admin.register(models.Service)
class ServiceAdmin(admin.ModelAdmin):
    list_display = ("name", "owner")
    list_filter = (("owner", admin.RelatedOnlyFieldListFilter),)
    list_select_related = ("owner",)


@admin.register(models.Project)
class ProjectAdmin(admin.ModelAdmin):
    list_display = ("name", "shard", "service", "farm", "owner")
    list_select_related = ("service", "farm", "shard", "owner")
    list_filter = ("shard", ("owner", admin.RelatedOnlyFieldListFilter))


class SenderForm(forms.ModelForm):
    sender = forms.ChoiceField(
        choices=[(entry.module_name, entry.module_name) for entry in plugins.notifications()]
    )

    class Meta:
        model = models.Sender
        exclude = ["content_object"]


@admin.register(models.Sender)
class SenderAdmin(admin.ModelAdmin):
    list_display = ("content_object", "content_type", "sender", "show_value", "owner")
    form = SenderForm
    list_filter = ("sender", "content_type")
    list_select_related = ("content_type",)
    inlines = [FilterInline]


@admin.register(models.Farm)
class FarmAdmin(admin.ModelAdmin):
    list_display = ("name", "source")
    list_filter = ("source",)


@admin.register(models.Exporter)
class ExporterAdmin(admin.ModelAdmin):
    list_display = ("job", "port", "path", "project", "enabled")
    list_filter = ("job", "port")
    readonly_fields = ("project",)


@admin.register(models.DefaultExporter)
class DefaultExporterAdmin(admin.ModelAdmin):
    list_display = ("job", "port", "path")
    list_filter = ("job", "port")


@admin.register(models.Probe)
class ProbeAdmin(admin.ModelAdmin):
    list_display = ("module", "description")


@admin.register(models.URL)
class URLAdmin(admin.ModelAdmin):
    # Disable add permission and project editing because of the difficult UI
    # but leave support for editing url/probe through admin panel
    def has_add_permission(self, request):
        return False

    list_display = ("url", "probe", "project")
    list_filter = ("probe", ("project__service", admin.RelatedOnlyFieldListFilter))
    list_select_related = ("project", "project__service", "probe")
    readonly_fields = ("project",)


@admin.register(models.Rule)
class RuleAdmin(admin.ModelAdmin):
    list_display = ("name", "clause", "duration", "content_object")
    list_filter = ("duration",)
    list_select_related = ("content_type",)

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.prefetch_related("content_object")


@admin.register(models.Prometheus)
class PrometheusAdmin(admin.ModelAdmin):
    list_display = ("shard", "host", "port")
    list_filter = ("shard",)
    actions = [
        actions.prometheus_targets,
        actions.prometheus_rules,
        actions.prometheus_urls,
        actions.prometheus_reload,
        actions.prometheus_tombstones,
    ]


@admin.register(models.Alert)
class AlertAdmin(admin.ModelAdmin):
    def __getattr__(self, name):
        # Override __getattr__ so that we can return a label
        # for any of our special values in list_display
        def __get_label(label):
            def __wrapped(instance):
                try:
                    return instance.json["commonLabels"][label]
                except KeyError:
                    return ""

            # We give the wrapped function the same description as
            # our label so that it shows up right in the admin panel
            __wrapped.short_description = label
            return __wrapped

        if name in self.list_display:
            return __get_label(name)

    date_hierarchy = "created"
    list_display = (
        "created",
        "datasource",
        "alertname",
        "service",
        "project",
        "severity",
        "job",
    )

    fields = ("created", "_json")
    readonly_fields = ("created", "_json")
    ordering = ("-created",)

    @admin.display(description="json")
    def _json(self, instance):
        return format_html("<pre>{}</pre>", json.dumps(instance.json, indent=2))

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import json
from django.urls import reverse
import yaml

from django.contrib.auth.models import Permission, User
from django.test import TestCase
from django.conf import settings


class Data:
    def __init__(self, *args, test_dir=settings.BASE_DIR / "promgen" / "tests"):
        self.path = test_dir.joinpath(*args)

    def json(self):
        with self.path.open() as fp:
            return json.load(fp)

    def yaml(self):
        with self.path.open() as fp:
            return yaml.safe_load(fp)

    def raw(self):
        with self.path.open() as fp:
            return fp.read()


class PromgenTest(TestCase):
    longMessage = True
    maxDiff = None
    fixtures = ["testcases.yaml"]

    def fireAlert(self, source="alertmanager.json", data=None):
        if data is None:
            data = Data("examples", source).raw()

        return self.client.post(reverse("alert"), data=data, content_type="application/json")

    def assertRoute(self, response, view, status=200, msg=None):
        self.assertEqual(response.status_code, status, msg)
        self.assertEqual(response.resolver_match.func.__name__, view.as_view().__name__)

    def assertCount(self, model, count, msg=None):
        self.assertEqual(model.objects.count(), count, msg)

    def force_login(self, username, **kwargs):
        user = User.objects.get(username=username, **kwargs)
        self.client.force_login(user, "django.contrib.auth.backends.ModelBackend")
        return user

    def assertMockCalls(self, mock_func, *args, **kwargs):
        # normally we would want to use mock_func.assert_called_with but
        # the check is too strict when it comes to dictionary arguments. By writing
        # our own wrapper around it, we can check the arguments match without
        # also requering the dictionary to have the same order
        # We split them here, instead of passing to assertEquals, so that we have
        # the arguments directly and not a call() object
        call_args, call_kwargs = mock_func.call_args
        self.assertEqual((call_args, call_kwargs), (args, kwargs))

    def add_user_permissions(self, *args, user=None):
        codenames = [p.split(".")[1] for p in args]
        permissions = Permission.objects.filter(
            codename__in=codenames, content_type__app_label="promgen"
        )

        if user is None:
            user = self.user

        user.user_permissions.add(*[p for p in permissions])


SETTINGS = Data("examples", "promgen.yml").yaml()



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import datetime
from unittest import mock

from django.test import override_settings
from django.urls import reverse

from promgen import forms, tests, errors

TEST_SETTINGS = tests.Data("examples", "promgen.yml").yaml()
TEST_DURATION = tests.Data("examples", "silence.duration.json").json()
TEST_RANGE = tests.Data("examples", "silence.range.json").json()

# Explicitly set a timezone for our test to try to catch conversion errors
TEST_SETTINGS["timezone"] = "Asia/Tokyo"


class SilenceTest(tests.PromgenTest):
    fixtures = ["testcases.yaml", "extras.yaml"]

    def setUp(self):
        self.user = self.force_login(username="demo")

    @override_settings(PROMGEN=TEST_SETTINGS)
    @mock.patch("promgen.util.post")
    def test_duration(self, mock_post):
        mock_post.return_value.status_code = 200

        with mock.patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.datetime(2017, 12, 14, tzinfo=datetime.timezone.utc)
            # I would prefer to be able to test with multiple labels, but since
            # it's difficult to test a list of dictionaries (the order is non-
            # deterministic) we just test with a single label for now
            self.client.post(
                reverse("proxy-silence"),
                data={
                    "duration": "1m",
                    "labels": {"instance": "example.com:[0-9]*"},
                },
                content_type="application/json",
            )
        mock_post.assert_called_with(
            "http://alertmanager:9093/api/v2/silences", json=TEST_DURATION
        )

    @override_settings(PROMGEN=TEST_SETTINGS)
    @mock.patch("promgen.util.post")
    def test_range(self, mock_post):
        mock_post.return_value.status_code = 200

        with mock.patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.datetime(2017, 12, 14, tzinfo=datetime.timezone.utc)
            self.client.post(
                reverse("proxy-silence"),
                data={
                    "startsAt": "2017-12-14 00:01",
                    "endsAt": "2017-12-14 00:05",
                    "labels": {"instance": "example.com:[0-9]*"},
                },
                content_type="application/json",
            )

        mock_post.assert_called_with(
            "http://alertmanager:9093/api/v2/silences", json=TEST_RANGE
        )

    @override_settings(PROMGEN=TEST_SETTINGS)
    def test_site_silence_errors(self):
        form = forms.SilenceForm(data={"labels": {}, "duration": "1m"})
        self.assertEqual(
            form.errors.as_data(),
            {"__all__": [errors.SilenceError.NOLABEL.error()]},
        )

        form = forms.SilenceForm(data={"labels": {"alertname": "example-rule"}, "duration": "1m"})
        self.assertEqual(
            form.errors.as_data(),
            {"__all__": [errors.SilenceError.GLOBALSILENCE.error()]},
        )

        form = forms.SilenceForm(
            data={"labels": {"alertname": "example-rule", "foo": "bar"}, "duration": "1m"}
        )
        self.assertEqual(
            form.errors.as_data(),
            {"__all__": [errors.SilenceError.GLOBALSILENCE.error()]},
        )

        form = forms.SilenceForm(
            data={"labels": {"alertname": "example-rule", "service": "foo"}, "duration": "1m"}
        )
        self.assertEqual(form.errors, {}, "Expected no errors")
        self.assertEqual(
            form.cleaned_data,
            {
                "comment": "Silenced from Promgen",
                "createdBy": "Promgen",
                "duration": "1m",
                "endsAt": "",
                "labels": {"alertname": "example-rule", "service": "foo"},
                "startsAt": "",
            },
        )



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging

from pkg_resources import working_set

logger = logging.getLogger(__name__)


def discovery():
    return working_set.iter_entry_points("promgen.discovery")


def notifications():
    return working_set.iter_entry_points("promgen.notification")


# Since plugins may need to load other resources bundled with them, we loop
# through an additional promgen.apps entry point so that the default django
# project loaders work as expected. This also should simplfy some configuration
# for plugin authors
apps_from_setuptools = [
    entry.module_name for entry in working_set.iter_entry_points("promgen.apps")
]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging

from django.core.management.base import BaseCommand

from promgen import prometheus, tasks

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--reload",
            action="store_true",
            help="Trigger Prometheus Reload",
        )
        parser.add_argument(
            "--mode",
            type=lambda x: int(x, 8),
            default=0o644,
            help="Set mode for output file (default 644)",
        )
        parser.add_argument(
            "out",
            nargs="?",
            help="Optionally specify an output file to use an atomic write operation",
        )

    def handle(self, **kwargs):
        if kwargs["out"]:
            tasks.write_config(kwargs["out"], kwargs["reload"], kwargs["mode"])
        else:
            self.stdout.write(prometheus.render_config())



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging
import textwrap

from django import forms
from django.template.loader import render_to_string

from promgen import models, plugins, util

logger = logging.getLogger(__name__)


def load(name):
    for driver in plugins.notifications():
        if name == driver.module_name:
            return driver.load()()
    raise ImportError("Unknown notification plugin %s" % name)


class FormSenderBase(forms.Form):
    value = forms.CharField(required=True)
    alias = forms.CharField(required=False)


class NotificationBase:
    """
    Base Notification class
    """

    form = FormSenderBase

    @classmethod
    def create(cls, **kwargs) -> models.Sender:
        return models.Sender.objects.create(sender=cls.__module__, **kwargs)

    @classmethod
    def help(cls):
        if cls.__doc__:
            return textwrap.dedent(cls.__doc__)

    def _send(self, target, alert):
        """
        Sender specific implementation

        This function will receive some kind of target value, such as an email
        address or post endpoint and an individual alert combined with some
        additional alert meta data
        """
        raise NotImplementedError()

    def config(self, key, default=KeyError):
        """
        Plugin specific configuration

        This wraps our PROMGEN settings so that a plugin author does not need to
        be concerned with how the configuration files are handled but only about
        the specific key.
        """
        try:
            return util.setting(key, default=default, domain=self.__module__)
        except KeyError as e:
            raise KeyError(f"Missing key for domain: {self.__module__} {key}") from e

    def render(self, template, context):
        s = render_to_string(template, context).strip()
        # Uncomment to re-generate test templates
        # with open(template.replace('sender', 'tests/notification'), 'w+') as fp:
        #     fp.write(s)
        return s



# Generated by Django 2.0.7 on 2018-07-31 08:09

from django.db import migrations
from django.conf import settings


def create_group(apps, schema_editor):
    if not settings.PROMGEN_DEFAULT_GROUP:
        return

    # Create Default Group
    group, created = apps.get_model('auth', 'Group').objects.get_or_create(
        name=settings.PROMGEN_DEFAULT_GROUP
    )

    # Create default permissions. We skip the permissions that are
    # generally for admin rules (Shards, Prometheus, Audit) and skip
    # custom permissions for label/annotations but list everything else
    Permission = apps.get_model('auth', 'Permission')
    group.permissions.set(
        Permission.objects.filter(
            content_type__app_label='promgen',
            content_type__model__in=[
                'exporter',
                'farm',
                'host',
                'project',
                'rule',
                'sender',
                'service',
                'url',
            ],
        )
    )

    # Add users to default group
    User = apps.get_model('auth', 'User')
    for user in User.objects.all():
        user.groups.add(group)


def remove_group(apps, schema_editor):
    apps.get_model('auth', 'Group').objects.filter(
        name=settings.PROMGEN_DEFAULT_GROUP
    ).delete()


class Migration(migrations.Migration):

    dependencies = [('promgen', '0002_auto_20180316_0525')]

    operations = [migrations.RunPython(create_group, remove_group)]




# Copyright (c) 2019 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from rest_framework import permissions, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from django.http import HttpResponse
from django.views.generic import View

from promgen import filters, models, prometheus, renderers, serializers, tasks


class AlertReceiver(View):
    def post(self, request, *args, **kwargs):
        # Normally it would be more 'correct' to check our 'alert_blacklist' here and avoid
        # writing to the database, but to keep the alert ingestion queue as simple as possible
        # we will go ahead and write all alerts to the database and then filter out (delete)
        # when we run tasks.process_alert
        alert = models.Alert.objects.create(body=request.body.decode("utf-8"))
        tasks.process_alert.delay(alert.pk)
        return HttpResponse("OK", status=202)


class AllViewSet(viewsets.ViewSet):
    permission_classes = [permissions.AllowAny]

    @action(detail=False, methods=["get"], renderer_classes=[renderers.RuleRenderer])
    def rules(self, request):
        rules = models.Rule.objects.filter(enabled=True)
        return Response(
            serializers.AlertRuleSerializer(rules, many=True).data,
            headers={"Content-Disposition": "attachment; filename=alert.rule.yml"},
        )


class ShardViewSet(viewsets.ModelViewSet):
    queryset = models.Shard.objects.all()
    filterset_class = filters.ShardFilter
    serializer_class = serializers.ShardSerializer
    lookup_field = "name"

    @action(detail=True, methods=["get"])
    def services(self, request, name):
        shard = self.get_object()
        return Response(serializers.ServiceSerializer(shard.service_set.all(), many=True).data)


class RuleMixin:
    @action(detail=True, methods=["get"], renderer_classes=[renderers.RuleRenderer])
    def rules(self, request, name):
        rules = models.Rule.objects.filter(obj=self.get_object())
        return Response(
            serializers.AlertRuleSerializer(rules, many=True).data,
            headers={"Content-Disposition": "attachment; filename=%s.rule.yml" % name},
        )


class NotifierMixin:
    @action(detail=True, methods=["get"])
    def notifiers(self, request, name):
        return Response(
            serializers.SenderSerializer(self.get_object().notifiers.all(), many=True).data
        )


class ServiceViewSet(NotifierMixin, RuleMixin, viewsets.ModelViewSet):
    queryset = models.Service.objects.all()
    filterset_class = filters.ServiceFilter
    serializer_class = serializers.ServiceSerializer
    lookup_value_regex = "[^/]+"
    lookup_field = "name"

    @action(detail=True, methods=["get"])
    def projects(self, request, name):
        service = self.get_object()
        return Response(serializers.ProjectSerializer(service.project_set.all(), many=True).data)

    @action(detail=True, methods=["get"])
    def targets(self, request, name):
        return HttpResponse(
            prometheus.render_config(service=self.get_object()),
            content_type="application/json",
        )


class ProjectViewSet(NotifierMixin, RuleMixin, viewsets.ModelViewSet):
    queryset = models.Project.objects.prefetch_related("service", "shard", "farm")
    filterset_class = filters.ProjectFilter
    serializer_class = serializers.ProjectSerializer
    lookup_value_regex = "[^/]+"
    lookup_field = "name"

    @action(detail=True, methods=["get"])
    def targets(self, request, name):
        return HttpResponse(
            prometheus.render_config(project=self.get_object()),
            content_type="application/json",
        )



# Copyright (c) 2020 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE


from django.core.exceptions import ValidationError

from promgen import models, validators
from promgen.tests import PromgenTest


class ModelTest(PromgenTest):
    def setUp(self):
        self.user = self.force_login(username="demo")

    def test_names(self):
        # Unicode is ok
        models.Service(name=r"日本語", owner=self.user).full_clean()
        # Spaces are ok
        models.Service(name=r"foo bar", owner=self.user).full_clean()
        # dash or under score are ok
        models.Service(name=r"foo-bar_baz", owner=self.user).full_clean()
        with self.assertRaises(ValidationError):
            # Fail a name with \
            models.Service(name=r"foo/bar", owner=self.user).full_clean()
            models.Service(name=r"foo\bar", owner=self.user).full_clean()

    def test_validators(self):
        with self.assertRaises(ValidationError, msg="Javascript injection"):
            validators.metricname(
                "asdasd[[1-1]]')) || (this.$el.ownerDocument.defaultView.alert('1337",
            )

        with self.assertRaises(ValidationError, msg="Vue.js injection"):
            validators.metricname(
                "[[this.$el.ownerDocument.defaultView.alert(1337)]]",
            )



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

"""
Promgen middleware

The middleware ensures three main things

1. We globally set request.site so that we can easily use it when searching for
   our global rule_set object

2. We store request.user globally so that we can retrieve it when logging users
   to our AuditLog

3. Since many different actions can trigger a write of the target.json or rules
files, we need to handle some deduplication. This is handled by using the django
caching system to set a key and then triggering the actual event from middleware
"""

import logging
from threading import local

from django.contrib import messages
from django.db.models import prefetch_related_objects

from promgen import models
from promgen.signals import trigger_write_config, trigger_write_rules, trigger_write_urls

logger = logging.getLogger(__name__)


_user = local()


class PromgenMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # This works the same as the django middleware
        # django.contrib.sites.middleware.CurrentSiteMiddleware
        # but ensures that it uses our proxy object so that test cases
        # properly find our rule_set object
        request.site = models.Site.objects.get_current()
        # Prefetch our rule_set as needed, since request.site is used on
        # many different pages
        prefetch_related_objects([request.site], "rule_set")

        # Get our logged in user to use with our audit logging plugin
        if request.user.is_authenticated:
            _user.value = request.user

        response = self.get_response(request)

        triggers = {
            "Config": trigger_write_config.send,
            "Rules": trigger_write_rules.send,
            "URLs": trigger_write_urls.send,
        }

        for msg, func in triggers.items():
            for receiver, status in func(self, request=request, force=True):
                if status is False:
                    messages.warning(request, "Error queueing %s " % msg)
        return response


def get_current_user():
    return getattr(_user, "value", None)



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from unittest import mock

from django.test import override_settings

from promgen import models, rest, tests
from promgen.notification.linenotify import NotificationLineNotify


class LineNotifyTest(tests.PromgenTest):
    def setUp(self):
        one = models.Project.objects.get(pk=1)
        two = models.Service.objects.get(pk=2)

        NotificationLineNotify.create(obj=one, value="hogehoge")
        NotificationLineNotify.create(obj=two, value="asdfasdf")

    @override_settings(PROMGEN=tests.SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    @mock.patch("promgen.util.post")
    def test_line_notify(self, mock_post):
        response = self.fireAlert()
        self.assertRoute(response, rest.AlertReceiver, 202)

        # Swap the status to test our resolved alert
        SAMPLE = tests.Data("examples", "alertmanager.json").json()
        SAMPLE["status"] = "resolved"
        SAMPLE["commonLabels"]["service"] = "other-service"
        SAMPLE["commonLabels"].pop("project")
        response = self.fireAlert(data=SAMPLE)
        self.assertRoute(response, rest.AlertReceiver, 202)

        _MESSAGE = tests.Data("notification", "linenotify.body.txt").raw().strip()
        _RESOLVED = tests.Data("notification", "linenotify.resolved.txt").raw().strip()

        mock_post.assert_has_calls(
            [
                mock.call(
                    "https://notify.example",
                    data={"message": _MESSAGE},
                    headers={"Authorization": "Bearer hogehoge"},
                ),
                mock.call(
                    "https://notify.example",
                    data={"message": _RESOLVED},
                    headers={"Authorization": "Bearer asdfasdf"},
                ),
            ],
            any_order=True,
        )



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging

from django import forms
from django.core.mail import send_mail

from promgen.notification import NotificationBase

logger = logging.getLogger(__name__)


class FormEmail(forms.Form):
    value = forms.CharField(
        required=True,
        label="Email Address",
    )
    alias = forms.CharField(
        required=False,
        help_text="Use to hide email from being displayed.",
    )


class NotificationEmail(NotificationBase):
    """
    Simple plain text email notification.
    """

    form = FormEmail

    def _send(self, address, data):
        subject = self.render("promgen/sender/email.subject.txt", data)
        body = self.render("promgen/sender/email.body.txt", data)
        send_mail(
            subject,
            body,
            self.config("sender"),
            [address],
        )
        return True



import os
import pathlib

from django.conf import settings
from django.core import checks

from promgen import util


@checks.register("settings")
def directories(**kwargs):
    for key in ["prometheus:rules", "prometheus:blackbox", "prometheus:targets"]:
        try:
            path = pathlib.Path(util.setting(key)).parent
        except TypeError:
            yield checks.Warning(
                f"Missing setting for {key} in {settings.PROMGEN_CONFIG_FILE} ",
                id="promgen.W001",
            )
        else:
            if not os.access(path, os.W_OK):
                yield checks.Warning("Unable to write to %s" % path, id="promgen.W002")


@checks.register("settings")
def promtool(**kwargs):
    key = "prometheus:promtool"
    try:
        path = pathlib.Path(util.setting(key))
    except TypeError:
        yield checks.Warning(
            f"Missing setting for {key} in {settings.PROMGEN_CONFIG_FILE} ",
            id="promgen.W001",
        )
    else:
        if not os.access(path, os.X_OK):
            yield checks.Warning("Unable to execute file %s" % path, id="promgen.W003")



# Copyright (c) 2019 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import yaml
from rest_framework import renderers


# https://www.django-rest-framework.org/api-guide/renderers/#custom-renderers
class RuleRenderer(renderers.BaseRenderer):
    format = "yaml"
    media_type = "application/x-yaml"
    charset = "utf-8"

    def render(self, data, media_type=None, renderer_context=None):
        return yaml.safe_dump(
            {"groups": [{"name": name, "rules": data[name]} for name in data]},
            default_flow_style=False,
            allow_unicode=True,
            encoding=self.charset,
        )



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging

from django import forms
from django.contrib.auth.models import User
from promgen import models
from promgen.notification import NotificationBase

logger = logging.getLogger(__name__)


def _choices():
    for u in User.objects.filter(is_active=True).order_by("username"):
        if u.first_name:
            yield (u.username, f"{u.username} ({u.first_name} {u.last_name})")
        elif u.email:
            yield (u.username, f"{u.username} ({u.email})")
        else:
            yield (u.username, u.username)


class FormUser(forms.Form):
    value = forms.ChoiceField(
        required=True,
        label="Username",
        choices=_choices,
    )


class NotificationUser(NotificationBase):
    """
    Send a notification to a specific user.
    """

    form = FormUser

    def splay(self, address):
        try:
            user = User.objects.get(username=address)
        except User.DoesNotExist:
            logger.error("Missing user %s", address)
        else:
            yield from models.Sender.objects.filter(obj=user)

    def _send(self, address, data):
        user = User.objects.get(username=address)
        for sender in models.Sender.objects.filter(obj=user, enabled=True):
            try:
                sender.driver._send(sender.value, data)
            except Exception:
                logger.exception("Error sending with %s", sender)
        return True



from django.contrib import admin, messages

from promgen import tasks


@admin.action(description="Clear Tombstones")
def prometheus_tombstones(modeladmin, request, queryset):
    for server in queryset:
        tasks.clear_tombstones.apply_async(queue=server.host)
        messages.info(request, "Clear Tombstones on " + server.host)


@admin.action(description="Reload Configuration")
def prometheus_reload(modeladmin, request, queryset):
    for server in queryset:
        tasks.reload_prometheus.apply_async(queue=server.host)
        messages.info(request, "Reloading configuration on " + server.host)


@admin.action(description="Deploy Prometheus Targets")
def prometheus_targets(modeladmin, request, queryset):
    for server in queryset:
        tasks.write_config.apply_async(queue=server.host)
        messages.info(request, "Deploying targets to " + server.host)


@admin.action(description="Deploy Prometheus Rules")
def prometheus_rules(modeladmin, request, queryset):
    for server in queryset:
        tasks.write_rules.apply_async(queue=server.host)
        messages.info(request, "Deploying rules to " + server.host)


@admin.action(description="Deploy Prometheus Urls")
def prometheus_urls(modeladmin, request, queryset):
    for server in queryset:
        tasks.write_urls.apply_async(queue=server.host)
        messages.info(request, "Deploying urls to " + server.host)


@admin.action(description="Deploy Datasource Targets")
def shard_targets(modeladmin, request, queryset):
    for shard in queryset:
        prometheus_targets(modeladmin, request, shard.prometheus_set.all())


@admin.action(description="Deploy Datasource Rules")
def shard_rules(modeladmin, request, queryset):
    for shard in queryset:
        prometheus_rules(modeladmin, request, shard.prometheus_set.all())


@admin.action(description="Deploy Datasource Urls")
def shard_urls(modeladmin, request, queryset):
    for shard in queryset:
        prometheus_urls(modeladmin, request, shard.prometheus_set.all())



# Generated by Django 2.2.8 on 2020-02-21 06:03

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0015_internal'),
    ]

    operations = [
        migrations.CreateModel(
            name='AlertLabel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('value', models.CharField(max_length=128)),
                ('alert', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Alert')),
            ],
        ),
    ]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

"""promgen URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.10/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from rest_framework import routers

from django.conf.urls import url
from django.contrib import admin
from django.urls import include, path
from django.views.decorators.csrf import csrf_exempt

from promgen import proxy, rest, views

router = routers.DefaultRouter()
router.register("all", rest.AllViewSet, basename="all")
router.register("service", rest.ServiceViewSet)
router.register("shard", rest.ShardViewSet)
router.register("project", rest.ProjectViewSet)


urlpatterns = [
    path("admin/", admin.site.urls),
    path("", views.HomeList.as_view(), name="home"),
    # Data source
    path("datasource", views.DatasourceList.as_view(), name="datasource-list"),
    path("datasource/<int:pk>", views.DatasourceDetail.as_view(), name="datasource-detail"),
    # Services
    path("new/service", views.ServiceRegister.as_view(), name="service-new"),
    path("service", views.ServiceList.as_view(), name="service-list"),
    path("service/<int:pk>", views.ServiceDetail.as_view(), name="service-detail"),
    path("service/<int:pk>/delete", views.ServiceDelete.as_view(), name="service-delete"),
    path("service/<int:pk>/new", views.ProjectRegister.as_view(), name="project-new"),
    path("service/<int:pk>/update", views.ServiceUpdate.as_view(), name="service-update"),
    path("service/<int:pk>/notifier", views.ServiceNotifierRegister.as_view(), name="service-notifier"),
    # Projects
    path("project/<int:pk>", views.ProjectDetail.as_view(), name="project-detail"),
    path("project/<int:pk>/delete", views.ProjectDelete.as_view(), name="project-delete"),
    path("project/<int:pk>/update", views.ProjectUpdate.as_view(), name="project-update"),
    path("project/<int:pk>/unlink", views.UnlinkFarm.as_view(), name="farm-unlink"),
    path("project/<int:pk>/link/<source>", views.FarmLink.as_view(), name="farm-link"),
    path("project/<int:pk>/newfarm", views.FarmRegister.as_view(), name="farm-new"),
    path("project/<int:pk>/exporter", views.ExporterRegister.as_view(), name="project-exporter"),
    path("project/<int:pk>/notifier", views.ProjectNotifierRegister.as_view(), name="project-notifier"),
    path("project/<int:pk>/scrape", views.ExporterScrape.as_view(), name="exporter-scrape"),
    # Exporters
    path("exporter/<int:pk>/delete", views.ExporterDelete.as_view(), name="exporter-delete"),
    path("exporter/<int:pk>/toggle", views.ExporterToggle.as_view(), name="exporter-toggle"),
    # URLs
    path("url", views.URLList.as_view(), name="url-list"),
    path("url/<int:pk>/new", views.URLRegister.as_view(), name="url-new"),
    path("url/<int:pk>/delete", views.URLDelete.as_view(), name="url-delete"),
    # Farms
    path("farm", views.FarmList.as_view(), name="farm-list"),
    path("farm/<int:pk>", views.FarmDetail.as_view(), name="farm-detail"),
    path("farm/<int:pk>/refresh", views.FarmRefresh.as_view(), name="farm-refresh"),
    path("farm/<int:pk>/hosts", views.HostRegister.as_view(), name="hosts-add"),
    path("farm/<int:pk>/update", views.FarmUpdate.as_view(), name="farm-update"),
    path("farm/<int:pk>/delete", views.FarmDelete.as_view(), name="farm-delete"),
    path("farm/<int:pk>/convert", views.FarmConvert.as_view(), name="farm-convert"),
    # Hosts
    path("host/", views.HostList.as_view(), name="host-list"),
    path("host/<slug>", views.HostDetail.as_view(), name="host-detail"),
    path("host/<int:pk>/delete", views.HostDelete.as_view(), name="host-delete"),
    # Notifiers
    path("notifier/<int:pk>/delete", views.NotifierDelete.as_view(), name="notifier-delete"),
    path("notifier/<int:pk>/test", views.NotifierTest.as_view(), name="notifier-test"),
    path("notifier/<int:pk>", views.NotifierUpdate.as_view(), name="notifier-edit"),
    path("notifier/<int:pk>/toggle", views.NotifierToggle.as_view(), name="notifier-toggle"),
    # Rules
    path("rule", views.RulesList.as_view(), name="rules-list"),
    path("rule/<int:pk>", views.RuleDetail.as_view(), name="rule-detail"),
    path("rule/<int:pk>/edit", views.RuleUpdate.as_view(), name="rule-edit"),
    path("rule/<int:pk>/delete", views.RuleDelete.as_view(), name="rule-delete"),
    path("rule/<int:pk>/toggle", views.RuleToggle.as_view(), name="rule-toggle"),
    path("rule/<int:pk>/test", csrf_exempt(views.RuleTest.as_view()), name="rule-test"),
    path("rule/<int:pk>/duplicate", views.RulesCopy.as_view(), name="rule-overwrite"),
    # Generic Rules
    path("<content_type>/<object_id>/rule", views.AlertRuleRegister.as_view(), name="rule-new"),
    # Other miscellaneous
    path("audit", views.AuditList.as_view(), name="audit-list"),
    path("site", views.SiteDetail.as_view(), name="site-detail"),
    path("profile", views.Profile.as_view(), name="profile"),
    path("import", views.Import.as_view(), name="import"),
    path("import/rules", views.RuleImport.as_view(), name="rule-import"),
    path("search", views.Search.as_view(), name="search"),
    path("metrics", csrf_exempt(views.Metrics.as_view()), name="metrics"),
    # Alerts
    path("alert", views.AlertList.as_view(), name="alert-list"),
    path("alert/<int:pk>", views.AlertDetail.as_view(), name="alert-detail"),
    # Third Party / Auth
    url("", include("django.contrib.auth.urls")),
    url("", include("social_django.urls", namespace="social")),
    # Legacy API
    path("api/v1/config", csrf_exempt(views.ApiConfig.as_view())),
    path("api/", include((router.urls, "api"), namespace="old-api")),
    path("api/v1/rules", csrf_exempt(views.RulesConfig.as_view())),
    path("api/v1/targets", csrf_exempt(views.ApiConfig.as_view()), name="config-targets"),
    path("api/v1/urls", csrf_exempt(views.URLConfig.as_view()), name="config-urls"),
    path("api/v1/alerts", csrf_exempt(rest.AlertReceiver.as_view()), name="alert"),
    path("api/v1/host/<slug>", views.HostDetail.as_view()),
    # Prometheus Proxy
    # these apis need to match the same path because Promgen can pretend to be a Prometheus API
    path("graph", proxy.ProxyGraph.as_view()),
    path("api/v1/labels", proxy.ProxyLabels.as_view(), name="proxy-label"),
    path("api/v1/label/<label>/values", proxy.ProxyLabelValues.as_view(), name="proxy-values"),
    path("api/v1/query_range", proxy.ProxyQueryRange.as_view()),
    path("api/v1/query", proxy.ProxyQuery.as_view(), name="proxy-query"),
    path("api/v1/series", proxy.ProxySeries.as_view()),
    # Alertmanager Proxy
    # Promgen does not pretend to be an Alertmanager so these can be slightly different
    path("proxy/v1/alerts", csrf_exempt(proxy.ProxyAlerts.as_view()), name="proxy-alerts"),
    path("proxy/v1/silences", csrf_exempt(proxy.ProxySilences.as_view()), name="proxy-silence"),
    path("proxy/v1/silences/<silence_id>", csrf_exempt(proxy.ProxyDeleteSilence.as_view()), name="proxy-silence-delete"),
    # Promgen rest API
    path("rest/", include((router.urls, "api"), namespace="api")),
]

try:
    import debug_toolbar

    urlpatterns += [
        path("__debug__/", include(debug_toolbar.urls)),
    ]
except ImportError:
    pass



# Copyright (c) 2019 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from django.contrib import messages
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.contrib.auth.views import redirect_to_login
from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404
from django.views.generic.base import ContextMixin

from promgen import models


class ContentTypeMixin:
    def set_object(self, content_type, object_id):
        self.content_type = ContentType.objects.get(model=content_type, app_label="promgen")
        self.object_id = object_id


class RuleFormMixin:
    # When creating a single object, we want to use the
    # default form class and delegate to form_valid but
    # when we are importing multiple objects, we delegate
    # a form_import class to handle processing
    def post(self, request, content_type, object_id):
        single = self.get_form(self.form_class)
        # Set an instance of our content_object here so that we can
        # pass it along for promtool to render
        single.instance.set_object(content_type, object_id)
        if single.is_valid():
            return self.form_valid(single)

        importer = self.get_form(self.form_import_class)
        if importer.is_valid():
            ct = ContentType.objects.get_by_natural_key("promgen", content_type).model_class()
            content_object = ct.objects.get(pk=object_id)

            return self.form_import(importer, content_object)

        return self.form_invalid(single)


class PromgenPermissionMixin(PermissionRequiredMixin):
    def handle_no_permission(self):
        messages.warning(self.request, self.get_permission_denied_message())
        return redirect_to_login(
            self.request.get_full_path(),
            self.get_login_url(),
            self.get_redirect_field_name(),
        )


class ShardMixin(ContextMixin):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if "pk" in self.kwargs:
            context["object"] = context["shard"] = get_object_or_404(
                models.Shard, id=self.kwargs["pk"]
            )
        return context


class ProjectMixin(ContextMixin):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if "pk" in self.kwargs:
            context["object"] = context["project"] = get_object_or_404(
                models.Project, id=self.kwargs["pk"]
            )
        return context


class ServiceMixin(ContextMixin):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if "pk" in self.kwargs:
            context["object"] = context["service"] = get_object_or_404(
                models.Service, id=self.kwargs["pk"]
            )
        return context



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging

from django.shortcuts import get_object_or_404

from promgen import models
from promgen import discovery

logger = logging.getLogger(__name__)


class DiscoveryPromgen(discovery.DiscoveryBase):
    """Promgen local database discovery plugin

    This is the default discovery plugin for farms and hosts stored locally in
    promgen's database. They are queried directly from Django's ORM
    """

    remote = False

    def fetch(self, farm_name):
        """Fetch list of hosts for a farm from the local database"""
        farm = get_object_or_404(models.Farm, name=farm_name)
        for host in models.Host.objects.filter(farm=farm):
            yield host.name

    def farms(self):
        """Fetch farms from local database"""
        for farm in models.Farm.objects.filter(source=discovery.FARM_DEFAULT):
            yield farm.name



# Generated by Django 1.11.10 on 2018-03-16 05:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0001_squashed_0044_common-rules'),
    ]

    operations = [
        migrations.AddField(
            model_name='shard',
            name='enabled',
            field=models.BooleanField(default=True, help_text='Able to register new Services and Projects'),
        ),
        migrations.AlterField(
            model_name='shard',
            name='proxy',
            field=models.BooleanField(default=False, help_text='Queries can be proxied to these shards'),
        ),
    ]



# Copyright (c) 2019 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from django.core.management.base import BaseCommand, CommandError

from promgen import models, util


class Command(BaseCommand):
    def add_arguments(self, parser):
        help_text = util.help_text(models.Host)

        parser.add_argument(
            "project",
            type=util.cast(models.Project),
            help="Existing Project",
        )
        parser.add_argument("host", help=help_text("name"))
        # parser.add_argument("--enabled", default=False, action="store_true", help=help_text('enabled'))

    def handle(self, project, **kwargs):
        if project.farm is None:
            raise CommandError("Project currently not associated with a farm :%s" % project)

        host, created = project.farm.host_set.get_or_create(name=kwargs["host"])
        if created:
            print("Registered host", host)
        else:
            print("Found existing host", host)



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE
from unittest import mock

import requests

from django.test import override_settings
from django.urls import reverse

from promgen import models, tests, views

TEST_SETTINGS = tests.Data("examples", "promgen.yml").yaml()
TEST_IMPORT = tests.Data("examples", "import.json").raw()
TEST_REPLACE = tests.Data("examples", "replace.json").raw()


class RouteTests(tests.PromgenTest):
    def setUp(self):
        self.user = self.force_login(username="demo")

    @override_settings(PROMGEN=TEST_SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    @mock.patch("promgen.signals._trigger_write_config")
    @mock.patch("promgen.tasks.reload_prometheus")
    def test_import(self, mock_write, mock_reload):
        self.add_user_permissions(
            "promgen.change_rule", "promgen.change_site", "promgen.change_exporter"
        )
        response = self.client.post(reverse("import"), {"config": TEST_IMPORT})

        self.assertRoute(response, views.Import, 302, "Redirect to imported object")
        self.assertCount(models.Service, 3, "Import one service (Fixture has two services)")
        self.assertCount(models.Project, 4, "Import two projects")
        self.assertCount(models.Exporter, 2, "Import two exporters")
        self.assertCount(models.Host, 3, "Import three hosts")

    @override_settings(PROMGEN=TEST_SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    @mock.patch("promgen.signals._trigger_write_config")
    @mock.patch("promgen.tasks.reload_prometheus")
    def test_replace(self, mock_write, mock_reload):
        self.add_user_permissions(
            "promgen.change_rule", "promgen.change_site", "promgen.change_exporter"
        )

        response = self.client.post(reverse("import"), {"config": TEST_IMPORT})
        self.assertRoute(response, views.Import, 302, "Redirect to imported object")

        response = self.client.post(reverse("import"), {"config": TEST_REPLACE})
        self.assertRoute(response, views.Import, 302, "Redirect to imported object (2)")

        self.assertCount(models.Service, 3, "Import one service (Fixture has two services)")
        self.assertCount(models.Project, 4, "Import two projects (Fixture has 2 projectsa)")
        self.assertCount(models.Exporter, 2, "Import two exporters")
        self.assertCount(
            models.Farm, 4, "Original two farms and one new farm (fixture has one farm)"
        )
        self.assertCount(models.Host, 5, "Original 3 hosts and two new ones")

    @mock.patch("requests.get")
    def test_scrape(self, mock_get):
        shard = models.Shard.objects.create(name="test_scrape_shard")
        service = models.Service.objects.create(name="test_scrape_service")
        farm = models.Farm.objects.create(name="test_scrape_farm")
        farm.host_set.create(name="example.com")
        project = models.Project.objects.create(
            name="test_scrape_project", service=service, shard=shard, farm=farm
        )

        # Uses the scrape target as the key, and the POST body that should
        # result in that URL
        exporters = {
            "http://example.com:8000/metrics": {
                "target": "#exporterresult",
                "job": "foo",
                "port": 8000,
                "scheme": "http",
            },
            "https://example.com:8000/foo": {
                "target": "#exporterresult",
                "job": "foo",
                "port": 8000,
                "path": "/foo",
                "scheme": "https",
            },
        }

        for url, body in exporters.items():
            response = requests.Response()
            response.url = url
            response.status_code = 200
            mock_get.return_value = response

            # For each POST body, check to see that we generate and attempt to
            # scrape the correct URL
            response = self.client.post(reverse("exporter-scrape", kwargs={"pk": project.pk}), body)
            self.assertRoute(response, views.ExporterScrape, 200)
            self.assertEqual(mock_get.call_args[0][0], url)

    def test_failed_permission(self):
        # Test for redirect
        for request in [{"viewname": "rule-new", "args": ("site", 1)}]:
            response = self.client.get(reverse(**request))
            self.assertRoute(response, views.AlertRuleRegister, 302)
            self.assertTrue(response.url.startswith("/login"))

    def test_other_routes(self):
        self.add_user_permissions("promgen.add_rule", "promgen.change_site")
        for request in [{"viewname": "rule-new", "args": ("site", 1)}]:
            response = self.client.get(reverse(**request))
            self.assertRoute(response, views.AlertRuleRegister, 200)
